// Simplified access to localStorage with extra checks (it seems to be little bit more anal than necessary but I want to catch as many stupid bugs as possible)
"use strict";
// require: none

var DH = DH || {};

DH.storage = (function () {
    var self = {};

    self.keyExists = function (aKey) {
        // return true if key exists in storage
        if (typeof aKey !== 'string') {
            throw "DH.storage.readString key " + aKey + " is not string!";
        }
        return localStorage.hasOwnProperty(aKey);
    };

    self.erase = function (aKey) {
        // erase single key
        if (typeof aKey !== 'string') {
            throw "DH.storage.erase key " + aKey + " is not string!";
        }
        localStorage.removeItem(aKey);
    };

    self.size = function (aKey) {
        // return size of a key's value in bytes
        if (!localStorage.hasOwnProperty(aKey)) {
            return 0;
        }
        return localStorage.getItem(aKey).length;
    };

    self.sizeAll = function (aHuman) {
        // return size used by entire storage
        var keys = self.keys(), i, t = 0;
        for (i = 0; i < keys.length; i++) {
            t += self.size(keys[i]);
        }
        if (aHuman) {
            if (t > 1024) {
                return Math.ceil(t / 1024) + ' kB';
            }
            return t + ' B';
        }
        return t;
    };

    self.keys = function () {
        // return all keys
        var k, keys = [];
        for (k in localStorage) {
            if (localStorage.hasOwnProperty(k)) {
                keys.push(k);
            }
        }
        return keys;
    };

    self.eraseAll = function (aNothing) {
        // erase entire storage
        if (aNothing !== undefined) {
            throw "DH.storage.eraseAll does not require parameter, perhaps you wanted to call DH.storage.erase(key)";
        }
        localStorage.clear();
    };

    self.debug = function () {
        // return size occupied by each keys and first few bytes of data
        var i, keys = self.keys().sort(), s = [], c, t = 0;
        for (i = 0; i < keys.length; i++) {
            c = self.size(keys[i]);
            t += c;
            s.push(keys[i] + ': ' + c + ' B = ' + self.readString(keys[i], '').substr(0, 80) + '...');
        }
        s.push('Total size: ' + t + ' B (' + (t / 1000).toFixed(0) + ' kB)');
        return s.join('\n');
    };

    self.readString = function (aKey, aDefault) {
        // read string
        if (typeof aKey !== 'string') {
            throw "DH.storage.readString key " + aKey + " is not string!";
        }
        if ((aDefault !== undefined) && (typeof aDefault !== 'string')) {
            throw "DH.storage.readString default " + aDefault + " is not string nor undefined!";
        }
        if (localStorage.hasOwnProperty(aKey)) {
            return localStorage.getItem(aKey);
        }
        return aDefault;
    };

    self.writeString = function (aKey, aValue) {
        // write string
        if (typeof aKey !== 'string') {
            throw "DH.storage.writeString key " + aKey + " is not string!";
        }
        if ((aValue !== undefined) && (typeof aValue !== 'string')) {
            throw "DH.storage.writeString value " + aValue + " is not string nor undefined!";
        }
        localStorage.setItem(aKey, aValue);
    };

    self.readBoolean = function (aKey, aDefault) {
        // read true/false, undefined as default, everything else is default with warning
        var s = self.readString(aKey);
        // console.info(aKey, aDefault, s, typeof s);
        if (s === undefined) {
            return aDefault || false;
        }
        if ((s !== 'true') && (s !== 'false')) {
            console.warn('DH.storage.readBoolean: unusual boolean value "' + s + '" for "' + aKey + '", using default');
            return aDefault || false;
        }
        return s === 'true';
    };

    self.writeBoolean = function (aKey, aValue) {
        // write true/false
        if ((aValue !== true) && (aValue !== false)) {
            console.warn('DH.storage.writeBoolean: unusual boolean value "' + aValue + '" for "' + aKey + '", using false');
        }
        self.writeString(aKey, aValue === true ? 'true' : 'false');
    };

    self.readNumber = function (aKey, aDefault) {
        // read number, undefined as default, everything else is default with warning
        var s = self.readString(aKey), f;
        if (s === undefined) {
            return aDefault || 0;
        }
        f = parseFloat(s);
        if (isNaN(f)) {
            console.warn('DH.storage.readNumber: unusual number value "' + s + '" for "' + aKey + '", using default');
            return aDefault || 0;
        }
        return f;
    };

    self.writeNumber = function (aKey, aValue) {
        // write number
        if (typeof aValue !== 'number') {
            console.warn('DH.storage.writeNumber: unusual number value "' + aValue + '" for "' + aKey + '", using 0');
            self.writeString(aKey, '0');
        } else {
            self.writeString(aKey, aValue.toString());
        }
    };

    self.inc = function (aKey, aDefault) {
        // read number, increment it, write it back
        var i = self.readNumber(aKey, aDefault);
        i++;
        self.writeNumber(aKey, i);
        return i;
    };

    self.readObject = function (aKey, aDefault) {
        // read object, undefined as default, everything else is default with warning
        var s = self.readString(aKey), o;
        if (aDefault === undefined) {
            aDefault = {};
        }
        if (typeof aDefault !== 'object') {
            console.warn('DH.storage.readObject: default is not object in "' + aKey + '" but "' + aDefault + '", using {}');
            aDefault = {};
        }
        if (s === undefined) {
            return aDefault;
        }
        o = JSON.parse(s);
        if (typeof o !== 'object') {
            console.warn('DH.storage.readObject: unusual value "' + s + '" for "' + aKey + '", using default');
            return aDefault;
        }
        return o;
    };

    self.writeObject = function (aKey, aValue) {
        // write object
        if (typeof aValue !== 'object') {
            console.warn('DH.storage.writeObject: unusual object value "' + aValue + '" for "' + aKey + '", using {}');
            self.writeString(aKey, '{}');
        } else {
            self.writeString(aKey, JSON.stringify(aValue));
        }
    };

    self.readArray = function (aKey, aDefault) {
        // read array, undefined as default, everything else is default with warning
        var s = self.readString(aKey), o;
        if (aDefault === undefined) {
            aDefault = [];
        }
        if (!Array.isArray(aDefault)) {
            console.warn('DH.storage.readArray: default is not array in "' + aKey + '" but "' + aDefault + '", using []');
            aDefault = [];
        }
        if (s === undefined) {
            return aDefault;
        }
        o = JSON.parse(s);
        if (!Array.isArray(o)) {
            console.warn('DH.storage.readArray: unusual value "' + s + '" for "' + aKey + '", using default');
            return aDefault;
        }
        return o;
    };

    self.writeArray = function (aKey, aValue) {
        // write array
        if (!Array.isArray(aValue)) {
            console.warn('DH.storage.writeArray: unusual array value "' + aValue + '" for "' + aKey + '", using []');
            self.writeString(aKey, '[]');
        } else {
            self.writeString(aKey, JSON.stringify(aValue));
        }
    };

    return self;
}());

